<?php
namespace org;

use think\Db;
use think\Config;
use think\Session;
use think\Request;
use think\Loader;

/**
 * 权限认证类
 * 功能特性：
 * 1，是对规则进行认证，不是对节点进行认证。用户可以把节点当作规则名称实现对节点进行认证。
 *      $auth=new Auth();  $auth->check('规则名称','用户id')
 * 2，可以同时对多条规则进行认证，并设置多条规则的关系（or或者and）
 *      $auth=new Auth();  $auth->check('规则1,规则2','用户id','and')
 *      第三个参数为and时表示，用户需要同时具有规则1和规则2的权限。 当第三个参数为or时，表示用户值需要具备其中一个条件即可。默认为or
 * 3，一个用户可以属于多个用户组(vip_must_auth_group_access表 定义了用户所属用户组)。我们需要设置每个用户组拥有哪些规则(vip_must_auth_group 定义了用户组权限)
 * 4，支持规则表达式。
 *      在vip_must_auth_rule 表中定义一条规则时，如果type为1， note字段就可以定义规则表达式。 如定义{score}>5  and {score}<100
 * 表示用户的分数在5-100之间时这条规则才会通过。
 */
//数据库
/*
-- ----------------------------
-- Table structure for `vip_must_auth_group`
-- ----------------------------
DROP TABLE IF EXISTS `vip_must_auth_group`;
CREATE TABLE `vip_must_auth_group` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT COMMENT 'id',
  `title` varchar(100) NOT NULL COMMENT '标题',
  `status` tinyint(1) NOT NULL DEFAULT '1',
  `rules` varchar(255) NOT NULL COMMENT '权限规则id',
  PRIMARY KEY (`id`)
) ENGINE=MyISAM AUTO_INCREMENT=3 DEFAULT CHARSET=utf8 COMMENT='权限组表';

-- ----------------------------
-- Table structure for `vip_must_auth_group_access`
-- ----------------------------
DROP TABLE IF EXISTS `vip_must_auth_group_access`;
CREATE TABLE `vip_must_auth_group_access` (
  `id` smallint(6) NOT NULL AUTO_INCREMENT COMMENT 'id',
  `sid` smallint(8) unsigned NOT NULL COMMENT '系统权限用户id',
  `gid` smallint(8) unsigned NOT NULL COMMENT '权限规则id',
  PRIMARY KEY (`id`)
) ENGINE=MyISAM AUTO_INCREMENT=3 DEFAULT CHARSET=utf8 COMMENT='用户权限组规则表';

-- ----------------------------
-- Table structure for `vip_must_auth_rule`
-- ----------------------------
DROP TABLE IF EXISTS `vip_must_auth_rule`;
CREATE TABLE `vip_must_auth_rule` (
  `id` int(10) unsigned NOT NULL AUTO_INCREMENT COMMENT 'id',
  `name` varchar(20) NOT NULL COMMENT '节点名称',
  `url` varchar(100) NOT NULL COMMENT '链接url',
  `pid` smallint(5) unsigned NOT NULL COMMENT '父id',
  `type` tinyint(1) unsigned NOT NULL DEFAULT '1',
  `icon` varchar(50) DEFAULT NULL COMMENT '图标',
  `order` tinyint(4) unsigned NOT NULL COMMENT '排序',
  `note` char(100) DEFAULT NULL COMMENT '备注',
  `status` tinyint(1) NOT NULL DEFAULT '1' COMMENT '状态',
  PRIMARY KEY (`id`)
) ENGINE=MyISAM AUTO_INCREMENT=75 DEFAULT CHARSET=utf8 COMMENT='必要规则表';
 */

class Auth {
    /**
     * @var object 对象实例
     */
    protected static $instance;
    /**
     * 当前请求实例
     * @var Request
     */
    protected $request;

    //默认配置
    protected $config = [
        'auth_on'           => 1,                       // 权限开关
        'auth_type'         => 1,                       // 认证方式，1为实时认证；2为登录认证。
        'auth_group'        => 'must_auth_group',       // 用户组数据表名
        'auth_group_access' => 'must_auth_group_access',// 用户-用户组关系表
        'auth_rule'         => 'must_auth_rule',        // 权限规则表
        'auth_user'         => 'must_system_user',      // 用户信息表
    ];

    /**
     * 类架构函数
     * Auth constructor.
     */
    public function __construct(){
        //可设置配置项 auth, 此配置项为数组。
        if ($auth = Config::get('auth')) {
            $this->config = array_merge($this->config, $auth);
        }
        // 初始化request
        $this->request = Request::instance();
    }

    /**
     * 初始化
     * @access public
     * @param array $options 参数
     * @return \think\Request
     */
    public static function instance($options = []){
        if (is_null(self::$instance)) {
            self::$instance = new static($options);
        }
        return self::$instance;
    }

    /**
     * 检查权限
     * @param        $name     string|array  需要验证的规则列表,支持逗号分隔的权限规则或索引数组
     * @param        $sid      int           认证用户的id
     * @param int    $type     认证类型
     * @param string $mode     执行check的模式
     * @param string $relation 如果为 'or' 表示满足任一条规则即通过验证;如果为 'and'则表示需满足所有规则才能通过验证
     * @return bool               通过验证返回true;失败返回false
     */
    public function check($name, $sid, $type = 1, $mode = 'url', $relation = 'or'){
        if (!$this->config['auth_on']) {
            return true;
        }
        // 获取用户需要验证的所有有效规则列表
        $authList = $this->getAuthList($sid, $type);
        if (is_string($name)) {
            $name = strtolower($name);
            if (strpos($name, ',') !== false) {
                $name = explode(',', $name);
            } else {
                $name = [$name];
            }
        }
        $list = []; //保存验证通过的规则名
        if ('url' == $mode) {
            $REQUEST = unserialize(strtolower(serialize($this->request->param())));
        }
        foreach ($authList as $auth) {
            $query = preg_replace('/^.+\?/U', '', $auth);
            if ('url' == $mode && $query != $auth) {
                parse_str($query, $param); //解析规则中的param
                $intersect = array_intersect_assoc($REQUEST, $param);
                $auth      = preg_replace('/\?.*$/U', '', $auth);
                if (in_array($auth, $name) && $intersect == $param) {
                    //如果节点相符且url参数满足
                    $list[] = $auth;
                }
            } else {
                if (in_array($auth, $name)) {
                    $list[] = $auth;
                }
            }
        }
        if ('or' == $relation && !empty($list)) {
            return true;
        }
        $diff = array_diff($name, $list);
        if ('and' == $relation && empty($diff)) {
            return true;
        }

        return false;
    }

    /**
     * 根据用户id获取用户组,返回值为数组
     * @param  $sid int     用户id
     * @return array       用户所属的用户组 array(
     *              array('uid'=>'用户id','group_id'=>'用户组id','title'=>'用户组名称','rules'=>'用户组拥有的规则id,多个,号隔开'),
     *              ...)
     */
    public function getGroups($sid){
        static $groups = [];
        if (isset($groups[$sid])) {
            return $groups[$sid];
        }
        // 转换表名
        $auth_group_access = Loader::parseName($this->config['auth_group_access'], 1);
        $auth_group        = Loader::parseName($this->config['auth_group'], 1);
        // 执行查询
        $user_groups  = Db::view($auth_group_access, 'sid,gid')
            ->view($auth_group, 'title,rules', "{$auth_group_access}.gid={$auth_group}.id", 'LEFT')
            ->where("{$auth_group_access}.sid='{$sid}' and {$auth_group}.status='1'")
            ->select();
        $groups[$sid] = $user_groups ?: [];

        return $groups[$sid];
    }

    /**
     * 获得权限列表
     * @param integer $sid 用户id
     * @param integer $type
     * @return array
     */
    protected function getAuthList($sid, $type){
        static $_authList = []; //保存用户验证通过的权限列表
        $t = implode(',', (array)$type);
        if (isset($_authList[$sid . $t])) {
            return $_authList[$sid . $t];
        }
        if (2 == $this->config['auth_type'] && Session::has('_auth_list_' . $sid . $t)) {
            return Session::get('_auth_list_' . $sid . $t);
        }
        //读取用户所属用户组
        $groups = $this->getGroups($sid);
        $ids    = []; //保存用户所属用户组设置的所有权限规则id
        foreach ($groups as $g) {
            $ids = array_merge($ids, explode(',', trim($g['rules'], ',')));
        }
        $ids = array_unique($ids);
        if (empty($ids)) {
            $_authList[$sid . $t] = [];

            return [];
        }
        $map = [
            'id'   => ['in', $ids],
            'type' => $type
        ];
        //读取用户组所有权限规则
        $rules = Db::name($this->config['auth_rule'])->where($map)->field('note,name')->select();
        //循环规则，判断结果。
        $authList = []; //
        foreach ($rules as $rule) {
            if (!empty($rule['note'])) {
                //根据note进行验证
                $user    = $this->getUserInfo($sid); //获取用户信息,一维数组
                $command = preg_replace('/\{(\w*?)\}/', '$user[\'\\1\']', $rule['note']);
                @(eval('$note=(' . $command . ');'));
                if ($note) {
                    $authList[] = strtolower($rule['name']);
                }
            } else {
                //只要存在就记录
                $authList[] = strtolower($rule['name']);
            }
        }
        $_authList[$sid . $t] = $authList;
        if (2 == $this->config['auth_type']) {
            //规则列表结果保存到session
            Session::set('_auth_list_' . $sid . $t, $authList);
        }

        return array_unique($authList);
    }

    /**
     * 获得用户资料
     * @param $sid
     * @return mixed
     */
    protected function getUserInfo($sid){
        static $user_info = [];
        $user = Db::name($this->config['auth_user']);
        // 获取用户表主键
        $_pk = is_string($user->getPk()) ? $user->getPk() : 'uid';
        if (!isset($user_info[$sid])) {
            $user_info[$sid] = $user->where($_pk, $sid)->find();
        }

        return $user_info[$sid];
    }
}